﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using FluentAssertions;
using xReactor.Common;

namespace xReactor.Tests
{
    [TestClass]
    public class PropertySystemIntegrationTests : PropertyTestBase
    {
        [TestMethod]
        public void ReevaluatesTarget_WhenSourcePropertyChanges()
        {
            var room = CreateRoom();

            room.PeopleLimit.Should().Be(0, "room area is 0 - not set yet.");
            room.SquareMeters = 100;
            room.PeopleLimit.Should().Be(10, "people limit should have been updated " +
                "after SquareMeters property changes");
        }

        [TestMethod]
        public void ReevaluatesTarget_WhenSubPropertyChanges()
        {
            var room = CreateRoom();
            var person = new Person();
            room.Owner = person;

            room.OwnerAge.Should().Be(0, "age is not changed yet");
            person.Age = 20;
            room.OwnerAge.Should().Be(20, "age has already been updated");
        }

        [TestMethod]
        public void ReevaluatesTarget_WhenSourceCollectionCountChanges()
        {
            var room = CreateRoom();
            var chair = new Chair();
            var sofa = new Sofa();

            room.LargestPlaceToSit.Should().BeNull("no furniture has been added yet");
            room.Furniture.Add(chair);
            room.LargestPlaceToSit.Should().BeSameAs(chair, "added a 1-man chair");
            room.Furniture.Add(sofa);
            room.LargestPlaceToSit.Should().BeSameAs(sofa, "added a 4-man sofa");
        }

        [TestMethod]
        public void ReevaluatesTarget_WhenSourceCollectionElementChanges()
        {
            var room = CreateRoom();
            var chair = new Chair();
            var sofa = new Sofa();

            room.Furniture.Add(chair);
            room.Furniture.Add(sofa);
            room.LargestPlaceToSit.Should().BeSameAs(sofa, "now sofa is bigger");

            chair.NumSeats = 5; //hmm, a five-seat chair? 
            room.LargestPlaceToSit.Should().BeSameAs(chair, "now chair was made larger." +
                "This requires the system to trace all changes on each item in the " +
                "collection. ");
        }

        [TestMethod]
        public void DoesNotReevaluateExpression_WhenItemChangesAfterRemoval()
        {
            var room = CreateRoom();
            var chair = new Chair();
            var sofa = new Sofa();

            room.Furniture.Add(chair);
            room.Furniture.Add(sofa);
            room.LargestPlaceToSit.Should().BeSameAs(sofa, "now sofa is bigger");
            room.Furniture.Remove(chair);
            chair.NumSeats = 5;

            room.LargestPlaceToSit.Should().BeSameAs(sofa, "although chair was made larger," +
                "the chair has been removed from Furniture beforehand.");
        }

        #region Raw INPC tests

        [TestMethod]
        public void ReevaluatesTarget_WhenReactivePropertyIsBuiltOnINPCObject()
        {
            var productInfo = new ProductInfo() { IsAvailable = true };
            var product = new Product(productInfo);

            product.IsAvailable.Should().Be(true, "this is initial value in product info");
            productInfo.IsAvailable = false;
            product.IsAvailable.Should().Be(false, "value has been set anew via product info instance");
        }

        #endregion
    }
}
