﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using Wintellect.PowerCollections;

namespace xReactor
{
    public static class ExpressionHelper
    {
        public static string GetNameFromExpression<T>(Expression<Func<T>> nameExpression)
        {
            try
            {
                return ((MemberExpression)nameExpression.Body).Member.Name;
            }
            catch (Exception inner)
            {
                throw new ArgumentException("Must be an expression in form of \"()=>PropertyName\".",
                    "nameExpression", inner);
            }
        }

        private static Func<object> CompileMemberExpression(MemberExpression expr)
        {
            return Expression.Lambda(expr).Compile() as Func<object>;
        }

        private static Func<INotifyPropertyChanged> CompileMemberExpressionINPC(MemberExpression expr)
        {
            var compiled = Expression.Lambda(expr).Compile();
            return compiled as Func<INotifyPropertyChanged>;
        }

        public static UsedPropertyChain[] GetUsedProperties<T>(Expression<Func<T>> valueExpression)
        {
            var set = new Set<UsedPropertyChain>();
            GetUsedProperties(set, valueExpression);
            UsedPropertyChain[] array = set.ToArray();
            foreach (var up in array)
            {
                up.TopLevelParent = up;
                up.RaiseOnParentChanged();
            }
            return array;
        }

        private static void GetUsedProperties(Set<UsedPropertyChain> container,
            MemberExpression memberExpression, UsedSubproperty child)
        {
            if (memberExpression == null ||
                memberExpression.Member.MemberType != MemberTypes.Property)
                return;

            var parentMember = memberExpression.Expression as MemberExpression;
            if (parentMember != null)
            {
                Func<INotifyPropertyChanged> parentRetriever = CompileMemberExpressionINPC(parentMember);
                if (parentRetriever == null)
                {
                    //This happens when parentMember refers to a non-observable collection 
                    return;
                }
                var usedsub = new UsedSubproperty(memberExpression.Member.Name, parentRetriever, child);
                usedsub.TargetRetriever = CompileMemberExpression(memberExpression);
                GetUsedProperties(container, parentMember, usedsub);
            }
            else
            {
                if (memberExpression.Expression is ParameterExpression)
                {
                    //Presumably, we are within a nested lambda parameter block.
                    //Skip, just for now. If there's a nested lambda like "c=>c.property" it 
                    //is pretty complicated to know when to update.
                    //TODO: 
                    return;
                }

                var constant = memberExpression.Expression as ConstantExpression;
                if (constant != null)
                {
                    var notifiable = constant.Value as INotifyPropertyChanged;
                    if (notifiable != null)
                    {
                        string propertyName = memberExpression.Member.Name;
                        var chain = new UsedPropertyChain(notifiable, propertyName, child);
                        chain.TargetRetriever = CompileMemberExpression(memberExpression);
                        container.Add(chain);
                    }
                }
                else
                {
                    string msg = "Traversed expression tree to the leaves, but" +
                        "neither MemberExpression nor ConstantExpression was found. " +
                        "A member expression should finally retrieve a target (constant).";
                    throw new InvalidOperationException(msg);
                }
            }
        }

        private static void GetUsedProperties(Set<UsedPropertyChain> container,
            Expression expression)
        {
            if (expression == null)
                return; //e.g. in case of static method calls object is null

            var lambda = expression as LambdaExpression;
            if (lambda != null)
            {
                GetUsedProperties(container, lambda.Body);
                return;
            }

            var binary = expression as BinaryExpression;
            if (binary != null)
            {
                GetUsedProperties(container, binary.Left);
                GetUsedProperties(container, binary.Right);
                return;
            }

            var unary = expression as UnaryExpression;
            if (unary != null)
            {
                GetUsedProperties(container, unary.Operand);
                return;
            }

            var methodCall = expression as MethodCallExpression;
            if (methodCall != null)
            {
                GetUsedProperties(container, methodCall.Object);
                foreach (var arg in methodCall.Arguments)
                {
                    GetUsedProperties(container, arg);
                }
                return;
            }

            var conditional = expression as ConditionalExpression;
            if (conditional != null)
            {
                GetUsedProperties(container, conditional.Test);
                GetUsedProperties(container, conditional.IfFalse);
                GetUsedProperties(container, conditional.IfTrue);
                return;
            }

            var member = expression as MemberExpression;
            if (member != null)
            {
                GetUsedProperties(container, member, null);
                return;
            }

            if (expression is ConstantExpression)
            {
                //Not traceable, nothing to do
                return;
            }

            string msg = string.Format(
                    "Expressions of type {0} are not supported.", expression.GetType());
            throw new NotSupportedException(msg);
        }
    }
}
