﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Linq.Expressions;
using System.Reactive;
using System.Reactive.Linq;
using System.Text;
using System.Threading.Tasks;

namespace xReactor
{
    /// <summary>
    /// The heart of xReactor responsible for tracking
    /// property changes and invoking update logic.
    /// </summary>
    public class Reactor
    {
        public event Action<IProperty> RxPropertyChanged;

        public PropertyChangedEventHandler PropertyChangedHandler
        {
            get;
            private set;
        }

        public PropertyChangingEventHandler PropertyChangingHandler
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the target object for which the reactor is created.
        /// </summary>
        /// <value>The target object for which the reactor is created.</value>
        public object Target
        {
            get;
            private set;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="T:Reactor"/> class.
        /// </summary>
        public Reactor(object target, PropertyChangedEventHandler propertyChangedHandler,
            PropertyChangingEventHandler propertyChangingHandler)
        {
            if (target == null)
                throw new ArgumentNullException("target");

            this.Target = target;
            this.PropertyChangedHandler = propertyChangedHandler;
            this.PropertyChangingHandler = propertyChangingHandler;
        }

        public Property<T> Create<T>(string name, T defaultValue = default(T))
        {
            return new Property<T>(this, name) { Value = defaultValue };
        }

        public Property<T> Create<T>(string name, Func<T> valueFunc, params IProperty[] fromProperties)
        {
            var property = Create(name, valueFunc());
            WhenPropertiesChange(fromProperties).Subscribe(coll => property.Value = valueFunc());
            return property;
        }

        public Property<T> Create<T>(string name, Expression<Func<T>> valueExpression)
        {
            Func<T> compiled = valueExpression.Compile();

            //Get value from compiled now to provide the default value:
            var property = Create<T>(name, compiled());

            UsedPropertyChain[] properties = ExpressionHelper.GetUsedProperties(valueExpression);
            WhenAnyPropertyChanges(properties).Subscribe(unit => property.Value = compiled());
            
            return property;
        }

        public IObservable<T> WhenPropertyChanges<T>(IProperty property)
        {
            var stream = Observable.FromEvent<IProperty>(
                h => this.RxPropertyChanged += h,
                h => this.RxPropertyChanged -= h);

            return stream.Where(p => p == property).Select(p => (T)p.Value);
        }

        public IObservable<object> WhenPropertiesChange(params IProperty[] properties)
        {
            var stream = Observable.FromEvent<IProperty>(
                h => this.RxPropertyChanged += h,
                h => this.RxPropertyChanged -= h);

            return stream.Where(p => Array.IndexOf(properties, p) >= 0);
        }

        private IObservable<Unit> WhenAnyPropertyChanges(params UsedPropertyChain[] properties)
        {
            var changeStream = properties.Select(p =>
                Observable.FromEvent(
                h => p.Changed += h,
                h => p.Changed -= h)).Merge();
            return changeStream;            
        }

        internal void RaisePropertyChanged<T>(Property<T> property, T oldValue, T newValue)
        {
            if (property == null)
                throw new ArgumentNullException("property");
            if (property.Reactor != this)
                throw new ArgumentException("Specified property's owner must be this instance.", "property");

            var target = Target;

            var changing = PropertyChangingHandler;
            if (changing != null)
            {
                changing(target, new PropertyChangingEventArgs(property.Name));
            }

            var changed = PropertyChangedHandler;
            if (changed != null)
            {
                changed(target, new PropertyChangedEventArgs(property.Name));
            }

            var rxChanged = RxPropertyChanged;
            if (rxChanged != null)
            {
                rxChanged(property);
            }
        }
    }
}
