﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Linq;
using System.Reactive.Disposables;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace xReactor
{
    abstract public class UsePropertyBase
    {
        UsedPropertyChain topLevelParent;
        INotifyPropertyChanged lastParent;
        object lastTarget;

        List<INotifyPropertyChanged> tracedItems;

        public string Name
        {
            get;
            private set;
        }

        public UsedSubproperty Child
        {
            get;
            private set;
        }

        public UsedPropertyChain TopLevelParent
        {
            get { return topLevelParent; }
            set
            {
                topLevelParent = value;
                if (Child != null)
                    Child.TopLevelParent = value;
            }
        }

        public Func<object> TargetRetriever
        {
            get;
            set;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="T:UsePropertyBase"/> class.
        /// </summary>
        public UsePropertyBase(string name, UsedSubproperty child)
        {
            if (name == null)
                throw new ArgumentNullException("name");

            this.Child = child;
            this.Name = name;
            this.tracedItems = new List<INotifyPropertyChanged>();
        }

        abstract protected INotifyPropertyChanged GetCurrentParent();

        virtual protected object GetCurrentTarget()
        {
            var retriever = this.TargetRetriever;
            if (retriever != null)
            {
                return retriever();
            }
            else return null;
        }

        internal void RaiseOnParentChanged()
        {
            INotifyPropertyChanged newParent = GetCurrentParent();
            object newTarget = GetCurrentTarget();

            if (newParent != lastParent)
            {
                UnhookParent(lastParent);
                HookParent(newParent);

                lastParent = newParent;
            }

            if (newTarget != lastTarget)
            {
                UnhookTarget(lastTarget);
                HookTarget(newTarget);

                lastTarget = newTarget;
            }

            if (Child != null)
                Child.RaiseOnParentChanged();
        }

        private void HookTarget(object target)
        {
            var sourceCollection = target as INotifyCollectionChanged;
            if (sourceCollection != null)
                sourceCollection.CollectionChanged += sourceCollection_CollectionChanged;
        }

        private void UnhookTarget(object target)
        {
            var sourceCollection = target as INotifyCollectionChanged;
            if (sourceCollection != null)
                sourceCollection.CollectionChanged -= sourceCollection_CollectionChanged;
        }

        private void HookParent(INotifyPropertyChanged parent)
        {
            if (parent != null)
                parent.PropertyChanged += source_PropertyChanged;

        }

        void UnhookParent(INotifyPropertyChanged parent)
        {
            if (parent != null)
                parent.PropertyChanged -= source_PropertyChanged;
        }

        void UnhookAllItems()
        {
            if (tracedItems != null)
            {
                foreach (var item in tracedItems)
                {
                    item.PropertyChanged -= item_PropertyChanged;
                }
                tracedItems.Clear();
            }
        }

        private void source_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (ShouldRaiseChangeNotification(e.PropertyName))
            {
                //update child before reevaluating expression
                //-may save some cycles
                if (Child != null)
                    Child.RaiseOnParentChanged();
                
                RaiseSourceChanged();
            }
        }

        void sourceCollection_CollectionChanged(object sender,
            NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == NotifyCollectionChangedAction.Reset)
            {
                UnhookAllItems();
            }
            else if (e.OldItems != null) //OldItems is null for Reset
            {
                foreach (var item in e.OldItems.OfType<INotifyPropertyChanged>())
                {
                    item.PropertyChanged -= item_PropertyChanged;
                    tracedItems.Remove(item);
                }
            }

            if (e.NewItems != null)
            {
                foreach (var item in e.NewItems.OfType<INotifyPropertyChanged>())
                {
                    item.PropertyChanged += item_PropertyChanged;
                    tracedItems.Add(item);
                }
            }

            RaiseSourceChanged();
        }

        void item_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            RaiseSourceChanged();
        }

        private bool ShouldRaiseChangeNotification(string changedPropertyName)
        {
            //Always notify if the source is a collection. This allows expressions
            //like '()=>collection.First()' etc. The 'collection' reference does
            //not change, but items do.
            return PropertyNamesMatch(this.Name, changedPropertyName);// || IsCollection;
        }

        private bool PropertyNamesMatch(string tracedName, string changedName)
        {
            return string.IsNullOrEmpty(changedName) || tracedName == changedName;
        }

        /// <summary>
        /// Call when there is a need to reevaluate property 
        /// from lambda expression.
        /// </summary>
        /// <param name="source"></param>
        /// <param name="e"></param>
        protected void RaiseSourceChanged()
        {
            TopLevelParent.RaiseChainChanged();
        }
    }

    public class UsedSubproperty : UsePropertyBase
    {
        Func<INotifyPropertyChanged> parentRetriever;

        /// <summary>
        /// Initializes a new instance of the <see cref="T:UsedSubproperty"/> class.
        /// </summary>
        public UsedSubproperty(string name, Func<INotifyPropertyChanged> parentRetriever, UsedSubproperty child)
            : base(name, child)
        {
            if (parentRetriever == null)
                throw new ArgumentNullException("parentRetriever");

            this.parentRetriever = parentRetriever;
        }

        protected override INotifyPropertyChanged GetCurrentParent()
        {
            return parentRetriever();
        }
    }

    public class UsedPropertyChain : UsePropertyBase
    {
        public event Action Changed;

        public INotifyPropertyChanged Parent
        {
            get;
            private set;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="T:UsedProperty"/> class.
        /// </summary>
        public UsedPropertyChain(INotifyPropertyChanged parent, string name, UsedSubproperty child)
            : base(name, child)
        {
            if (parent == null)
                throw new ArgumentNullException("target");

            this.Parent = parent;
            this.TopLevelParent = this;
        }

        protected internal void RaiseChainChanged()
        {
            var handler = this.Changed;
            if (handler != null)
            {
                handler();
            }
        }

        protected override INotifyPropertyChanged GetCurrentParent()
        {
            return Parent;
        }
    }
}
